import std/[httpclient, json, jsonutils, logging, options, sets, strutils,
            terminal, times, tables]
import timeutils, uuids, zero_functional
import ./formatting, ./libpit

type
  PbmVsbSyncContext* = object
    apiBaseUrl*: string
    apiToken*: string
    issueContext*: string
    pit: PitConfig
    http: HttpClient

  ServerTask* = object
    id*: string
    summary*: string
    details*: string
    state*: string
    lastUpdatedAt*: DateTime
    archivedAt*: Option[DateTime]
    tags*: seq[string]
    parent*: Option[string]
    priority*: Option[string]
    project*: Option[string]
    milestone*: Option[string]
    hideUntil*: Option[DateTime]


proc `%`*(dt: DateTime): JsonNode = %(dt.formatIso8601)
func toJsonHook(dt: DateTime): JsonNode = %(dt.formatIso8601)
proc fromJsonHook(dt: var DateTime, n: JsonNode): void =
  dt = n.getStr.parseIso8601


#func `%`*(p: Project): JsonNode = toJson(p)
func `%`*(t: ServerTask): JsonNode = toJson(t)


proc toServerTask(i: Issue): ServerTask =
  return ServerTask(
    id: $i.id,
    summary: i.summary,
    details: i.details,
    state: $i.state,
    tags: i.tags,
    lastUpdatedAt:
      if i.hasProp("last-updated"): parseIso8601(i["last-updated"])
      else: now(),
    parent:
      if i.hasProp("parent"): some(i["parent"])
      else: none[string](),
    priority:
      if i.hasProp("priority"): some(i["priority"])
      else: none[string](),
    project:
      if i.hasProp("project"): some(i["project"])
      else: none[string](),
    milestone:
      if i.hasProp("milestone"): some(i["milestone"])
      else: none[string](),
    hideUntil:
      if i.hasProp("hide-until"): some(parseIso8601(i["hide-until"]))
      else: none[DateTime]())


func getOrFail(n: JsonNode, k: string): JsonNode =
  if not n.hasKey(k):
    raise newException(ValueError, "missing key '" & k & "'")
  return n[k]


proc initSyncContext*(pit: PitConfig, syncConfig: JsonNode): PbmVsbSyncContext =
  result.pit = pit
  result.apiBaseUrl = syncConfig.getOrFail("apiBaseUrl").getStr
  result.apiToken = syncConfig.getOrFail("apiToken").getStr
  result.issueContext = syncConfig.getOrFail("context").getStr
  result.http = newHttpClient()
  result.http.headers = newHttpHeaders({ "Authorization": "Bearer " & result.apiToken })


proc fetchServerTasks*(ctx: PbmVsbSyncContext): seq[ServerTask] =
  result = newSeq[ServerTask]()
  let url = ctx.apiBaseUrl & "/tasks"

  debug "Fetching tasks from server:\n\t" & url
  let resp = ctx.http.get(url)
  if resp.status != "200":
    debug("Received " & resp.status & ": " & resp.body)
    raise newException(Exception, "Failed to fetch tasks from server")
  fromJson(result, parseJson(resp.body)["data"])


proc updateTask*(ctx: PbmVsbSyncContext, task: ServerTask): void =
  let url = ctx.apiBaseUrl & "/task" & task.id
  let body = %task
  let resp = ctx.http.post(url, $body)
  if resp.status != "200":
    debug("Received " & resp.status & ": " & resp.body)
    raise newException(Exception, "Failed to update task " & task.id & "on server")


proc updateTasks*(ctx: PbmVsbSyncContext, tasks: seq[ServerTask]): void =
  let url = ctx.apiBaseUrl & "/tasks"
  let body = %tasks
  let resp = ctx.http.put(url, $body)
  if resp.status != "200":
    debug("Received " & resp.status & ": " & resp.body)
    raise newException(Exception, "Failed to update tasks on server")


proc deleteTask*(ctx: PbmVsbSyncContext, taskId: string): void =
  let url = ctx.apiBaseUrl & "/tasks/" & taskId
  let resp = ctx.http.delete(url)
  if resp.status != "200":
    debug("Received " & resp.status & ": " & resp.body)
    raise newException(Exception, "Failed to delete task " & taskId & " on server")


proc sync*(
    ctx: PbmVsbSyncContext,
    dryRun = true,
    batchSize = 100): void =

  # We're going to do a uni-directional sync, pushing local issues to the
  # server. However, we only want to update issues that have changed since
  # the last sync based on the *last-updated* property.

  # Note that all the logic assumes the server only has one context of tasks.
  # If this ever changes, this logic will need to change.
  let filter = propsFilter(newTable({ "context": ctx.issueContext }))
  let allIssues = ctx.pit.tasksDir.loadAllIssues()

  var issues = newSeq[Issue]()
  for state in allIssues.keys:
    issues.add(allIssues[state].filter(filter))

  let serverTasks = fetchServerTasks(ctx)
  debug "Loaded $# server tasks" % [$serverTasks.len]
  var issuesToPush = newSeq[ServerTask]()

  var unmatchedServerTaskIds = toHashSet(serverTasks --> map(it.id))

  # Process all local issues
  info "Updating the following tasks for context " & ctx.issueContext
  for lentIssue in issues:
    let i = lentIssue
    let foundTask = serverTasks --> find(it.id == $i.id)

    if foundTask.isSome:
      # There is a server task for this issue
      unmatchedServerTaskIds.excl(foundTask.get.id)

      if i.hasProp("last-updated"):
        var localUpdate = parseIso8601(i["last-updated"])
        localUpdate.nanosecond = 0
        if  foundTask.get.lastUpdatedAt >= localUpdate:
          # but we don't have any update
          continue

    # We fell through the conditional block above, so either there isn't a
    # server task, or we *do* have an update,
    issuesToPush.add(toServerTask(i))
    info "  " & formatSectionIssue(i, width = terminalWidth() - 8)

  # Now archive the issues on the server that we didn't see locally.
  info "Archiving the following tasks for context " & ctx.issueContext
  for task in serverTasks:
    var toArchive = task
    if unmatchedServerTaskIds.contains(task.id) and
       toArchive.archivedAt.isNone:
      toArchive.archivedAt = some(now())
      issuesToPush.add(toArchive)

  if not dryRun:
    var offset = 0
    while (offset < issuesToPush.len):
      let batchSize = min(issuesToPush.len - offset, batchSize)
      updateTasks(ctx, issuesToPush[offset ..< offset + batchSize])
      offset += batchSize
